#ifndef PARABOLICFEM2D_H
#define PARABOLICFEM2D_H

#include <vector>
#include "densematrix.h"
#include "triangulation.h"
#include "timedeppde2d.h"
#include "cg.h"
#include "dirichletbc2d.h"

/** ParabolicFEM1D:
 *      Provides structures for the assembly and solution 
 *      of a 2D instationary (symmetric) finite element problem
 *      on a given ordered mesh.
 */
class ParabolicFEM2D{

private:
    /* Private Member Variables */
    
    Triangulation& mesh;                            // The 2D triangular mesh
    std::vector<std::vector<double> > solution;     // The solution matrix: a vector at each time step
    DirichletBC2D& d_bc;                            // The object that specifies the Dirichlet boundary
    								                //  (has a function is_dirichlet_point(x))

    SparseMatrix A;                                 // System operator matrix A_h
    SparseMatrix M;                                 // Mass matrix M_h
    std::vector<double> F;                          // Right hand side vector b_h
    
    TimedepPDE2D& pde2d;                            // PDE Object specifying the problem to be solved
    
    int timesteps;                                  // Number of time steps in the theta scheme
    double T;                                       // Final time
    double theta;                                   // The theta of the theta-scheme
        
    /* Private Member Functions */
    
    /** Performs the assembly of M_h
     */
    void
    assemble_M();
   
    /** Performs the assembly of A_h and b_h
     */ 
    void
    assemble_A(int current_timestep);

    /** Fills the solution at time t = 0 with initial condition
     */
    void
    set_initial_condition();
    
public:
    
    /* Public Member Functions */
    
    /** Constructor
     */
    ParabolicFEM2D(Triangulation& _mesh, TimedepPDE2D& _pde2d,     
                   int _timesteps, double _T, double _theta,
                   DirichletBC2D& _d_bc);
            
    /** Assembles and solves the linear system
     */
    void
    solve();
    
    /** Prints both A_h and b_h to std::cout.
     *  Can be used to debug the system.
     */
    void
    write_system();
    
    /** Write the solution to the specified file
     *  (in format: x_1 x_2 u(x))
     *  or print the vector to std::cout if no filename is given
     */
    void
    write_solution(std::string filename = std::string());
    
    /** Write the solution to the specified file
     *  (in triangular mesh format)
     *  or print the vector to std::cout if no filename is given
     */
    void
    write_solution_on_mesh(std::string filename = std::string());

};

//===========================================
// IMPLEMENTATIONS
//===========================================

/********
 * TODO * 
 ********/
ParabolicFEM2D::ParabolicFEM2D(Triangulation& _mesh, TimedepPDE2D& _pde2d,     
               int _timesteps, double _T, double _theta,
                DirichletBC2D& _d_bc){}

/********
 * TODO * 
 ********/
void
ParabolicFEM2D::set_initial_condition()
{
    /* Fill the solution vector at t=0 with initial condition
     */
}
 
/********
 * TODO * 
 ********/   
void
ParabolicFEM2D::assemble_A(int current_timestep)
{
    /* Performs the element-wise assembly of A_h(t) and b_h(t) at the given time step.
     * (Note that we can assume that pde2d.set_time() has been called before,
     * so that the pde2d-object already has the correct time.)
     * Handles the Dirichlet boundary conditions. Depending on where you set u = w + g, 
     * you might need the current_timestep here.
     */
}

/********
 * TODO * 
 ********/
void
ParabolicFEM2D::assemble_M()
{
    /* Assembles the mass matrix M_h using
     * the transformation onto the reference element
     * for each triangle.
     */    
}

/********
 * TODO * 
 ********/
void
ParabolicFEM2D::solve(){
    
    /* Solves the system using a general theta scheme.
     */
}

void
ParabolicFEM2D::write_system()
{
    std::cout << "A = " << std::endl << A << std::endl;
    std::cout << "b = " << std::endl << F << std::endl;
}

void
ParabolicFEM2D::write_solution(std::string filename)
{
    if(filename.size()>0){
        std::ofstream out(filename.c_str());
        if(out.good()){
            double deltaT = T / (double)timesteps;
            for(int t = 0; t <= timesteps; ++t){
                for(int i = 0; i < mesh.get_n_nodes(); ++i){
                    out << t*deltaT << " " << mesh.get_node(i+1).x << " " << mesh.get_node(i+1).y << " " << solution[t][i] << std::endl;
                }
                out << std::endl;
            }
        }
        out.close();
        return;
    }

    for(int t = 0; t <= timesteps; ++t){
        std::cout << "u = " << std::endl << solution[t] << std::endl;
    }
}

void
ParabolicFEM2D::write_solution_on_mesh(std::string filename)
{
    if(filename.size()>0){
        std::ofstream out(filename.c_str());
        if(out.good()){
            double deltaT = T / (double)timesteps;
            for(int t = 0; t <= timesteps; ++t){
                for (int i=1; i<= mesh.get_n_elements(); i++) {
                    Polygon& p = mesh.get_element(i);
                    for(int j = 0; j < p.N; ++j){
                         out << t*deltaT << " " << mesh.get_node(p.nodes[j]).x << " " 
                                        << mesh.get_node(p.nodes[j]).y << " " << solution[t][p.nodes[j]-1] << std::endl;
                    }
                    out << t*deltaT << " " << mesh.get_node(p.nodes[0]).x << " " 
                                    << mesh.get_node(p.nodes[0]).y << " " << solution[t][p.nodes[0]-1] << std::endl;
                    out << std::endl << std::endl;
                }
            }
        }
        out.close();
        return;
    }
    
    double deltaT = T / (double)timesteps;
    for(int t = 0; t <= timesteps; ++t){
        for (int i=1; i<= mesh.get_n_elements(); i++) {
            Polygon& p = mesh.get_element(i);
            for(int j = 0; j < p.N; ++j){
                 std::cout << t*deltaT << " " << mesh.get_node(p.nodes[j]).x << " " 
                          << mesh.get_node(p.nodes[j]).y << " " << solution[t][p.nodes[j]-1] << std::endl;
            }
            std::cout << t*deltaT << " " << mesh.get_node(p.nodes[0]).x << " " 
                      << mesh.get_node(p.nodes[0]).y << " " << solution[t][p.nodes[0]-1] << std::endl;
            std::cout << std::endl << std::endl;
        }
    }
}

#endif // PARABOLICFEM2D_H