#ifndef POISSON2D_H
#define POISSON2D_H

#include "densematrix.h"
#include "timedeppde2d.h"

typedef double (*Fct2D)(Point2D&);
typedef double (*TimedepFct2D)(double t, Point2D& x);

/** Poisson2D:
 *      Problem class that implements a 2D Laplace Operator
 *      
 *      Optionally with 
 *       - (non-homogeneous) Dirichlet boundary conditions
 *       - (non-homogeneous) initial values for non-stationary problems
 */
class TimedepPoisson2D : public TimedepPDE2D{
    
    TimedepFct2D f;         // Right hand side function, now f = f(t,x)
    TimedepFct2D d_bc;      // Dirichlet boundary condition function, d_bc = d_bc(t,x)
    Fct2D u0_fct;           // Initial value condition function
    
    DenseMatrix K_1, K_2, K_3;   // The matrices storing the integrals over the derivatives
                                    // of the shape functions, constant on the reference element.
    
    /** Function to set the values of the K matrices
     */
    void
    initialize_K_matrizes();

public:
        
    TimedepPoisson2D(TimedepFct2D _f);
    TimedepPoisson2D(TimedepFct2D _f, TimedepFct2D _d_bc);
    TimedepPoisson2D(TimedepFct2D _f, TimedepFct2D _d_bc, Fct2D _u0_fct);
    
    void
    assemble_on_element(Point2D& x1, Point2D& x2, Point2D& x3, double Ak[3][3], double bk[3]);
    
    double
    dirichlet_bc(Point2D& x);

    double
    initial_condition(Point2D& x);

};

TimedepPoisson2D::TimedepPoisson2D(TimedepFct2D _f) : 
    f(_f), K_1(3), K_2(3), K_3(3)
{
    initialize_K_matrizes();
}

TimedepPoisson2D::TimedepPoisson2D(TimedepFct2D _f, TimedepFct2D _d_bc) : 
    f(_f), d_bc(_d_bc),K_1(3), K_2(3), K_3(3)
{
    initialize_K_matrizes();
}

TimedepPoisson2D::TimedepPoisson2D(TimedepFct2D _f, TimedepFct2D _d_bc, Fct2D _u0_fct) : 
    f(_f), d_bc(_d_bc), u0_fct(_u0_fct), K_1(3), K_2(3), K_3(3)
{
    initialize_K_matrizes();
}

/********
 * TODO * 
 ********/
void
TimedepPoisson2D::initialize_K_matrizes()
{
    /* Exactly the same as in Poisson2D 
     */
}

void
TimedepPoisson2D::assemble_on_element(Point2D& x1, Point2D& x2, Point2D& x3, double Ak[3][3], double bk[3])
{
    /* Fill A and b with the correct values.
     * Basically the same as in Poisson2D, but 
     * f has to be called with 2 arguments now.
     */
}

double
TimedepPoisson2D::dirichlet_bc(Point2D& x){
    return d_bc(current_time, x);
};

double
TimedepPoisson2D::initial_condition(Point2D& x){
    if(u0_fct != NULL){
        return u0_fct(x);
    }
    else{
        return 0;
    }
};

#endif // POISSON2D_H